@@ -34,11 +34,6 @@
 #include <QTimer>
 class QSocketNotifier;
 
-#if HAVE_FAM
-#include <fam.h>
-#include <limits.h>
-#endif
-
 #include <ctime>
 #include <sys/types.h> // time_t, ino_t
 
@@ -48,6 +43,10 @@ class QSocketNotifier;
 #include <QFileSystemWatcher>
 #endif // HAVE_QFILESYSTEMWATCHER
 
+#if HAVE_SYS_INOTIFY_H
+struct inotify_event;
+#endif
+
 /* KDirWatchPrivate is a singleton and does the watching
  * for every KDirWatch instance in the application.
  */
@@ -63,7 +62,6 @@ public:
         UnknownMode = 0,
         StatMode,
         INotifyMode,
-        FAMMode,
         QFSWatchMode,
     };
     enum {
@@ -152,11 +150,6 @@ public:
             });
         }
 
-#if HAVE_FAM
-        FAMRequest fr;
-        bool m_famReportedSeen;
-#endif
-
 #if HAVE_SYS_INOTIFY_H
         int wd;
         // Creation and Deletion of files happens infrequently, so
@@ -184,7 +177,6 @@ public:
     void startScan(KDirWatch *instance, bool notify, bool skippedToo);
 
     void removeEntries(KDirWatch *instance);
-    void statistics();
 
     void addWatch(Entry *entry);
     void removeWatch(Entry *entry);
@@ -194,18 +186,21 @@ public:
 
     static bool isNoisyFile(const char *filename);
 
-    void ref();
-    void unref();
+    void ref(KDirWatch *watch);
+    void unref(KDirWatch *watch);
+
+#if HAVE_SYS_INOTIFY_H
+    QString inotifyEventName(const inotify_event *event) const;
+#endif
 
 public Q_SLOTS:
     void slotRescan();
-    void famEventReceived(); // for FAM
     void inotifyEventReceived(); // for inotify
     void slotRemoveDelayed();
     void fswEventReceived(const QString &path); // for QFileSystemWatcher
 
 public:
-    QTimer timer;
+    QTimer m_statRescanTimer;
     EntryMap m_mapEntries;
 
     KDirWatch::Method m_preferredMethod, m_nfsPreferredMethod;
@@ -221,16 +216,6 @@ public:
     bool rescan_all;
     QTimer rescan_timer;
 
-#if HAVE_FAM
-    QSocketNotifier *sn;
-    FAMConnection fc;
-    bool use_fam;
-
-    void checkFAMEvent(FAMEvent *fe);
-    bool useFAM(Entry *e);
-    void disableFAM();
-#endif
-
 #if HAVE_SYS_INOTIFY_H
     QSocketNotifier *mSn;
     bool supports_inotify;
@@ -248,9 +233,10 @@ public:
 
 private:
     // Public objects that reference this thread-local private instance.
-    uint m_references;
+    QList<KDirWatch *> m_referencesObjects;
 };
 
+QDebug operator<<(QDebug debug, const KDirWatchPrivate &dwp);
 QDebug operator<<(QDebug debug, const KDirWatchPrivate::Entry &entry);
 
 #endif // KDIRWATCH_P_H
